
"""
Compute CO2/NO2 mass along line or within area.
"""

import warnings

from scipy.constants import N_A

from shapely.geometry import LineString, Polygon

import numpy as np
import scipy.optimize
import xarray

import doas
import amrs.vis

import emiss.vis
import smartcarb.map

from . import misc
from . import functions


warnings.filterwarnings('ignore', 'Mean of empty slice')


def compute_mass(data, curve, field):
    """
    """
    distances = np.arange(0, np.nanmax(data.xp), 2e3)

    mass, mass_std = integrate_fields(data, field, curve, distances,
                                      tmax=100e3)

    valids = (distances > dmin) & np.isfinite(fluxes) & np.isfinite(fluxes_std)
    valids &= (distances <= dmax)

    mask = ~valids

    return np.ma.array(mass, mask=mask), np.ma.array(mass_std, mask=mask)



def integrate_fields(data, tracer, curve, distances, tmax):

    mass = []
    mass_std = []

    for d in distances:
        f,u = integrate_field(data, tracer, curve, d, tmax=tmax, show=False)
        mass.append(f)
        mass_std.append(u)

    return np.array(fluxes), np.array(fluxes_std)



def integrate_along_line(data, tracer, distances, curve=None, show=False,
                         method='sub-areas'):
    """
    distance from source origin
    """
    masses_true = []
    masses = []
    masses_std = []

    wind_eff = []
    wind_true = []

    for i in range(distances.size-1):

        results = extract_line_density(data, tracer, distances[i],
                                       distances[i+1],
                                       ya=None, yb=None, dy=5e3,
                                       method=method, show=False)

        if method == 'two-gauss':
            masses.append(results['co2-gauss'])
            masses_std.append(results['co2-gauss_std'])
        else:
            masses.append(results[method])
            masses_std.append(results['%s_std' % method])

        masses_true.append(results['density_true'])

        wind_eff.append(results['wind'])
        wind_true.append(results['wind_true'])

    wind_true = np.array(wind_true)
    wind_eff = np.array(wind_eff)

    print(np.array(masses))

    return np.array(masses), np.array(masses_std), np.array(masses_true), \
           wind_eff, wind_true




def add_area(ax, data, curve, distances, delta=10e3, color='yellow',
             add_label=True, origin='Berlin'):

    # across-plume distance
    y = get_plume_width(data)

    # compute arc length for parameters: arc(t)
    t = np.linspace(0.0, curve.t_o + distances.max() + delta, 5001)
    arc = misc.compute_arc_length(curve, curve.t_o, t)

    for i in range(distances.size):

        # interpolate distance from arc(t)
        xmin, xmax = np.interp(
            [distances[i] - delta / 2, distances[i] + delta / 2],
            arc, t
        )

        # compute normals
        (x0, x1), (y0, y1) = curve.compute_normal(xmin, t=y)
        (x3, x2), (y3, y2) = curve.compute_normal(xmax, t=y)

        # plot box
        label = 'control polygons 'if add_label and (i == 0) else None
        xx,yy = np.array([x0, x1, x2, x3, x0]), np.array([y0, y1, y2, y3, y0])
        smartcarb.map.add_curve(ax, xx, yy, origin=origin, ls='-',
                                color=color, label=label)



def sort_and_remove_nans(y, c, c_std):
    """
    Sort for y and remove nans.
    """
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]
    c_std = c_std[sort]

    valids = np.isfinite(y) & np.isfinite(c)

    return y[valids], c[valids], c_std[valids]



def fit_two_gaussian(y1, c1, c_std1, y2, c2, c_std2):

    def co2_func(x, *a):
        """
        """
        q = a[0]
        sigma = a[1]
        mu = a[2]

        return functions.gauss(x, q, sigma, mu)


    def no2_func(x, *a):

        q = a[3]
        sigma = a[1] # use same sigma as for co2
        mu = a[4]

        return functions.gauss(x, q, sigma, mu)


    # sort and remove nans
    y1, c1, c_std1 = sort_and_remove_nans(y1, c1, c_std1)
    y2, c2, c_std2 = sort_and_remove_nans(y2, c2, c_std2)

    # concat
    y = np.concatenate([y1, y2])
    c = np.concatenate([c1, c2])
    c_std = np.concatenate([c_std1, c_std2])


    def cost_function(x, *a):
        return np.concatenate([co2_func(y1, *a), no2_func(y2, *a)])

    p0 = np.array([
        100.0,   # CO2 line density (kg m-1)
         10e3,   # sigma of CO2 and NO2 function (m)
         0.0,    # CO2 center shift
          0.1,   # NO2 line density (kg m-1)
          0.0    # NO2 center shift
    ])

    if y.size < p0.size:
        return {'co2-gauss': np.nan, 'co2-gauss_std': np.nan}


    if np.all(np.isnan(c_std)):
        sigma = None
    elif np.any(np.isnan(c_std)):
        raise ValueError
    else:
        sigma = c_std


    warnings.simplefilter("error", scipy.optimize.OptimizeWarning)
    try:
        p, cov_p = scipy.optimize.curve_fit(cost_function, y, c, p0, sigma=sigma,
                                            absolute_sigma=sigma is not None)

    except (scipy.optimize.OptimizeWarning, RuntimeError):
        return {'co2-gauss': np.nan, 'co2-gauss_std': np.nan}

    p_std = np.sqrt(cov_p.diagonal())

    results = {}

    # fits
    results['y_fit'] = np.linspace(y.min(), y.max(), 501)
    results['co2-fit'] = co2_func(results['y_fit'], *p)
    results['no2-fit'] = no2_func(results['y_fit'], *p)

    # CO2 line denisty
    results['co2-gauss_fit'] = p[0]
    results['co2-gauss'] = p[0]
    results['co2-gauss_std'] = np.nan if sigma is None else p_std[0]

    # NO2 line density
    results['no2-gauss'] = p[3]
    results['no2-gauss_std'] = np.nan if sigma is None else p_std[3]

    return results



def fit_function(y, c, c_std, s=None, function='gauss'):


    # sort
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]
    c_std = c_std[sort]

    # remove nans
    valids = np.isfinite(c)
    y = y[valids]
    c = c[valids]
    c_std = c_std[valids]

    if function in ['gaussian', 'gauss']:
        curve_function = functions.gauss
        p0 = np.array([
            100.0,  # integral
            10e3,   # sigma
            0.0     # shift
        ])

    elif function == 'erf':
        curve_function = functions.error_function
        p0 = np.array([
            100.0, # integral
            1e3,   # sigma
            20e3,  # box width
            0.0    # shift
        ])
    else:
        raise ValueError

    if np.all(np.isnan(c_std)):
        sigma = None
    else:
        sigma = c_std

    if y.size < p0.size:
        return np.nan, np.nan, np.full(s.shape, np.nan)

    try:
        p, cov_p = scipy.optimize.curve_fit(curve_function, y, c, p0, sigma=sigma,
                                            absolute_sigma=sigma is not None)
    except RuntimeError as e:
        return np.nan, np.nan, np.full(s.shape, np.nan)

    if function in ['gaussian', 'gauss',  'erf']:
        integral = p[0]
        integral_std = np.sqrt(cov_p[0,0])
    else:
        raise ValueError

    return integral, integral_std, curve_function(s, *p)



def standard_error_of_the_mean(x):

    x = np.asarray(x)
    n = np.isfinite(x).sum()

    if n > 0:
        x = np.nanmean(x)
        return x / np.sqrt(n)

    return np.nan


def extract_pixels(data, tracer, xa, xb, ya, yb, dy=None, get_wind=False):
    """
    Extract pixels within a polygon given by plume coords for along-plume
    direction [xa,xb] and across-plume direction [ya,yb] (units: meters).
    """
    polygon = xarray.Dataset()


    # only use pixels that are valid observations
    xp = data['xp'].values
    yp = data['yp'].values

    area = (xa <= xp) & (xp < xb) & (ya <= yp) & (yp <= yb)

    isfinite = data['%s_isfinite' % tracer[:3]].values[area]

    # pixel in area
    polygon['x'] = xarray.DataArray(data.xp.values[area], name='x',
                                    dims='pixels')
    polygon['y'] = xarray.DataArray(data.yp.values[area], name='y',
                                    dims='pixels')

    c = data[tracer].values[area]
    c[~isfinite] = np.nan
    p = data['detected_plume'].values[area]

    polygon['c'] = xarray.DataArray(c, name='c', dims='pixels')
    polygon['is_plume'] = xarray.DataArray(p, name='is_plume', dims='pixels')

    if get_wind:
        if 'wind_eff_%s' % tracer[:3] in data:
            wind = data['wind_eff_%s' % tracer[:3]].values[area]
        elif 'u_eff_%s' % tracer[:3] in data:
            ueff = data['u_eff_%s' % tracer[:3]].values[area]
            veff = data['v_eff_%s' % tracer[:3]].values[area]
            wind = np.sqrt(ueff**2 + veff**2)
        else:
            wind = np.full(isfinite.shape, np.nan)

        wind[~isfinite] = np.nan

        polygon['wind'] = xarray.DataArray(wind, name='wind', dims='pixels')



    # estimate noise
    noise_level = data[tracer].attrs.get('noise_level', np.nan)

    if isinstance(noise_level, str):
        molar_mass = 46.0055e-3 # of NO2
        noise_level = {
            'low':  1e15 * molar_mass / N_A * 1e4, # molec cm-2 -> kg m-2
            'high': 2e15 * molar_mass / N_A * 1e4,
            'S5':   1.3e15 * molar_mass / N_A * 1e4
        }[noise_level]

    c_std = np.full(np.shape(c), noise_level)
    polygon['c_std'] = xarray.DataArray(c_std, name='c_std', dims='pixels')


    polygon['subpolygons'] = xarray.DataArray(
        np.arange(ya + 0.5 * dy, yb, dy), dims='subpolygons'
    )

    sub_areas = [(y0 - 0.5 * dy <= polygon['y']) &
                 (polygon['y'] < y0 + 0.5 * dy)
                 for y0 in polygon['subpolygons']]

    xx = get_values_from_areas(polygon['x'], sub_areas)
    yy = get_values_from_areas(polygon['y'], sub_areas)

    c = get_values_from_areas(polygon['c'], sub_areas)
    c_std = get_values_from_areas(polygon['c_std'], sub_areas)

    for name in ['x', 'y', 'c', 'c_std', 'is_plume']:

        if name == 'is_plume':
            function = np.sum
        elif name == 'c_std':
            function = standard_error_of_the_mean
        else:
            function = np.nanmean

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', r'Mean of empty slice')
            values = [function(v) for v in get_values_from_areas(polygon[name].values, sub_areas)]

        polygon['%s_sub' % name] = xarray.DataArray(values, dims='subpolygons')

    #if get_wind: TODO
    #    wind = np.nanmean(wind) if np.any(np.isfinite(wind)) else np.nan

    return polygon




def get_values_from_areas(values, sub_areas):
    return np.array([values[a.values] if np.any(a) else np.array([]) for a in sub_areas])



def extract_true_line_density(data, tracer, xa, xb, ya, yb):
    """
    """
    # only use pixels that are valid observations
    xp = data['xp'].values
    yp = data['yp'].values

    # use area of true plume if larger than [ya,yb]
    area_true = (xa <= xp) & (xp < xb) & data['%s_true_plume' % tracer[:3]].values

    if np.any(area_true):
        #ya_true, yb_true = get_plume_width(data, dy=5e3, area=area_true)
        ya_true, yb_true = -40e3, 40e3
        ya = min(ya, ya_true)
        yb = max(yb, yb_true)

    # area
    area = (xa <= xp) & (xp < xb) & (ya <= yp) & (yp <= yb)

    # get wind field
    if 'wind_eff_%s' % tracer[:3] in data:
        wind = data['wind_eff_%s' % tracer[:3]].values[area]
    else:
        ueff = data['u_eff_%s' % tracer[:3]].values[area]
        veff = data['v_eff_%s' % tracer[:3]].values[area]
        wind = np.sqrt(ueff**2 + veff**2)

    # get data
    tracer = '%s_true_signal_mass' % tracer[:3]
    c = data[tracer].values[area]

    # include wind speeds at pixel using weights
    c = (c * wind) / np.nanmean(wind)

    line_density = np.sum(c * 2e3**2 / (xb -xa))


    # wind speed
    if 'wind_eff_%s' % tracer[:3] in data:
        wind = data['wind_eff_%s' % tracer[:3]].values[area]
        wind = np.nanmean(wind)

    else:
        ueff = data['u_eff_%s' % tracer[:3]].values[area]
        veff = data['v_eff_%s' % tracer[:3]].values[area]
        wind = np.nanmean(np.sqrt(ueff**2 + veff**2))

    return line_density, yp[area], c, wind


def get_plume_width(data, dy=5e3, area='detected_plume'):

    # distance from center line
    if isinstance(area, str):
        yp = data['yp'].values[data[area]]
    else:
        yp = data['yp'].values[area]

    ymin = np.floor((yp.min() - 2 * dy) / dy) * dy
    ymax = np.ceil((yp.max() + 2 * dy) / dy) * dy

    return ymin, ymax


def extract_line_density(data, tracer, xa, xb, ya=None, yb=None, dy=5e3,
                         method='both', show=False, show_true=True,
                         add_errors=False):
    """
    Extract line densitiy of tracer in data within polygon given by along-plume
    interval [xa, xb] and across-plume interval [ya, yb]. If ya or yb is None
    estimate plume width from detected pixels. The plume center line is
    described by the `curve`.

    The function uses two methods:
        - fitting a Gaussian curve
        - summing mean values in sub-areas in across-plume direction using
          sub-areas with width `dy`
    """
    if ya is None or yb is None:
        ya, yb = get_plume_width(data)

    # true line density
    if ('%s_true_plume' % tracer) in data:
        true_density, y_true, c_true, wind_true = extract_true_line_density(data, tracer, xa, xb, ya, yb)
    else:
        true_density = np.nan
        y_true = np.nan
        c_true = np.nan
        wind_true = np.nan

    # returns
    results = {
        'area': (
            (xa <= data.xp) & (data.xp < xb) &
            (ya <= data.yp) & (data.yp <= yb) &
            data['%s_isfinite' % tracer[:3]].values
        ),
        'density_true': true_density,
        'wind_true':    wind_true,
        'y_fit':        np.linspace(ya, yb, 501)
    }
    # extract pixels in area
    polygon = extract_pixels(data, tracer, xa, xb, ya, yb, dy=dy, get_wind=True)
    no2 = None


    if method in ['gauss', 'sub-areas', 'both']:

        if method in ['gauss', 'gaussian', 'both']:

            # fit function (Gaussian or Error function)
            mass_gauss, mass_gauss_std, fit_gauss = fit_function(polygon['y'],
                                                                 polygon['c'],
                                                                 polygon['c_std'],
                                                                 results['y_fit'],
                                                                 function='gauss')

            if mass_gauss < 0.0 or mass_gauss > 1000.0 or mass_gauss_std > 1000.0:
                print('Fit failed:', mass_gauss, mass_gauss_std)
                results['gauss'] = np.nan
                results['gauss_std'] = np.nan
            else:
                results['gauss'] = mass_gauss
                results['gauss_std'] = mass_gauss_std


            results['gauss_fit'] = fit_gauss

            # set gauss to nan if only missing values in some sub-polygons with
            # detections
            if np.any((polygon['is_plume_sub'] > 0) & np.isnan(polygon['c_sub'])):
                results['gauss'] = np.nan
                results['gauss_std'] = np.nan

        if method in ['sub-areas', 'both']:

            if np.any((polygon['is_plume_sub'] > 0) & np.isnan(polygon['c_sub'])):
                results['sub-areas'] = np.nan
                results['sub-areas_std'] = np.nan
            else:

                valids = np.isfinite(polygon['c_sub'].values)

                ss = polygon['subpolygons'].values
                area_means = polygon['c_sub'].values
                area_means_std = polygon['c_std_sub'].values

                if np.all(~valids):
                    area_means[:] = np.nan
                else:
                    area_means = np.interp(ss, ss[valids], area_means[valids],
                                           left=0.0, right=0.0)

                results['sub-areas'] = np.sum(area_means * dy)
                results['sub-areas_std'] = np.nansum(area_means_std * dy)


    elif method in ['two-gauss']:
        co2 = polygon
        no2 = extract_pixels(data, tracer.replace('CO2', 'NO2'), xa, xb,
                             ya, yb, dy=dy, get_wind=True)

        res = fit_two_gaussian(co2.y, co2.c, co2.c_std, no2.y, no2.c, no2.c_std)
        results.update(res)

        if np.any((polygon['is_plume_sub'] > 0) & np.isnan(polygon['c_sub'])):
            results['co2-gauss'] = np.nan
            results['co2-gauss_std'] = np.nan

        if (
            results['co2-gauss'] < 0.0 or results['co2-gauss'] > 1000.0 or
            results['co2-gauss_std'] > 1000.0
        ):
            print('Fit failed:', results['co2-gauss'], results['co2-gauss_std'])
            results['co2-gauss'] = np.nan
            results['co2-gauss_std'] = np.nan

    else:
        raise ValueError


    if 'wind' in polygon:
        results['wind'] = np.nanmean(polygon['wind'])

    results = xarray.Dataset(results)

    if show:
        emiss.vis.plot_across_section(polygon, results, tracer=tracer[:3],
                                      method=method, no2=no2,
                                      show_true=show_true,
                                      add_errors=add_errors)

    return results







